// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <png.h>
#include <unistd.h>
#include <fcntl.h>
#include <setjmp.h>
#include <math.h>

#include "logger.h"
#include "eedigits.h"

#define DAY (24 * 60 * 60)
#define HOUR (60 * 60)
int tzone;

const char *month_names[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
			     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

FILE *
openlog(time_t when)
{
  char buf[20];
  struct tm *tm = localtime(&when);
  sprintf(buf, "%04d/%02d/%02d", tm->tm_year + 1900, tm->tm_mon+1, tm->tm_mday);
  FILE *f = fopen(buf, "rb");
  return f;
}

int max_events = 0;
int num_events = 0;
LogEvent *events = 0;

void
snarf_log_events(FILE *f)
{
  while (1)
    {
      if (num_events == max_events)
	{
	  max_events += 300;
	  events = (LogEvent *) realloc (events, max_events * sizeof(LogEvent));
	}
      int n = fread (events + num_events, sizeof(LogEvent), max_events - num_events, f);
      if (n <= 0)
	return;
      num_events += n;
    }
}

int
log_sort (const void *va, const void *vb)
{
  time_t a = ((LogEvent *)va)->time;
  time_t b = ((LogEvent *)vb)->time;
  return a - b;
}

void
zone(int z, int t, int h, int v)
{
  printf("[%d %4.1fF %4.1f%%%s] ",
	 z, t/100.0, h/2.0, v ? " Z" : " -");
}

#define font_height 8
#define font_width 6
int width = 600;
int height = 400;
int ndays = 1;
int bdays = 0;

png_byte **rgb;

static void
set_rgb (int x, int y, int r, int g, int b)
{
  png_byte *p = rgb[y] + x*3;
  p[0] = r;
  p[1] = g;
  p[2] = b;
}

static inline void
set_r (int x, int y, int r)
{
  rgb[y][x*3] = r;
}

static inline void
set_g (int x, int y, int g)
{
  rgb[y][x*3+1] = g;
}

static inline void
set_b (int x, int y, int b)
{
  rgb[y][x*3+2] = b;
}

static inline void
set_k (int x, int y, int k)
{
  set_rgb(x, y, k, k, k);
}

static inline void
set_p (int x, int y, int p)
{
  set_rgb (x, y, (p>>16)&0xff, (p>>8)&0xff, p&0xff);
}

static inline void
set_op (int x, int y, int p)
{
  png_byte *bp = rgb[y] + x*3;
  bp[0] |= (p>>16)&0xff;
  bp[1] |= (p>>8)&0xff;
  bp[2] |= p&0xff;
}

// x,y is lower left corner
// 'r' is rotation CLOCKWISE
static void
draw_char (int x, int y, int r, char c, int p)
{
  int cx, cy;
  unsigned const char *fontp = eedigits[c < 112 ? 0 : 1];
  if (c >= 112) c -= 112; else c -= 32;
  fontp += 6*c;
  switch (r)
    {
    case 0:
      y += 7;
      break;
    case 1:
      x += 7;
      break;
    case 2:
      y -= 7;
      break;
    case 3:
      x -= 7;
      break;
    }
  for (cx=0; cx<6; cx++)
    for (cy=0; cy<8; cy++)
      {
	if (fontp[cx] & (1<<cy))
	  {
	    switch (r)
	      {
	      case 0:
		set_p (x+cx, y-cy, p);
		break;
	      case 1:
		set_p (x-cy, y-cx, p);
		break;
	      case 2:
		set_p (x-cx, y+cy, p);
		break;
	      case 3:
		set_p (x+cy, y+cx, p);
		break;
	      }
	  }
      }
}

static void
draw_string (int x, int y, int r, char *s, int p)
{
  while (*s)
    {
      draw_char (x, y, r, *s++, p);
      switch (r)
	{
	case 0:
	  x += 6;
	  break;
	case 1:
	  y -= 6;
	  break;
	case 2:
	  x -= 6;
	  break;
	case 3:
	  y += 6;
	  break;
	}
    }
}

#define swap(a,b) { a ^= b; b ^= a; a ^= b; }

static void
draw_line (int x1, int y1, int x2, int y2, int p)
{
  if (x1 == x2)
    {
      if (y2 < y1)
	swap(y2, y1);
      while (y1 <= y2)
	set_p (x1, y1++, p);
    }
  else if (y1 == y2)
    {
      if (x2 < x1)
	swap(x2, x1);
      png_byte *bp = rgb[y1]+x1*3;
      png_byte *be = bp + 3;
      int i = (x2-x1) * 3;
      set_p (x1, y1, p);
      while (i--)
	*be++ = *bp++;
    }
  else
    {
      double step, v;
      step = abs(x1-x2);
      v = abs(y1-y2);
      if (step < v)
	step = v;
      for (v=0; v<=1; v+=1/step)
	{
	  double x = (x2-x1)*v+x1;
	  double y = (y2-y1)*v+y1;
	  set_p ((int)x, (int)y, p);
	}
    }
}

static void
draw_oline (int x1, int y1, int x2, int y2, int p)
{
  if (x1 == x2)
    {
      if (y2 < y1)
	swap(y2, y1);
      while (y1 <= y2)
	set_op (x1, y1++, p);
    }
  else if (y1 == y2)
    {
      if (x2 < x1)
	swap(x2, x1);
      while (x1 <= x2)
	set_op (x1++, y1, p);
    }
  else
    {
      double step, v;
      step = abs(x1-x2);
      v = abs(y1-y2);
      if (step < v)
	step = v;
      for (v=0; v<=1; v+=1/step)
	{
	  double x = (x2-x1)*v+x1;
	  double y = (y2-y1)*v+y1;
	  set_op ((int)(x+0.5), (int)(y+0.5), p);
	}
    }
}

static void
fill (int x1, int y1, int x2, int y2, int p)
{
  while (y1 <= y2)
    {
      draw_line (x1, y1, x2, y1, p);
      y1 ++;
    }
}

/* ====================================================================== */

time_t
daystart(time_t t)
{
  t -= (t-tzone) % DAY;
  return t;
}

#define GMARGIN 1
#define TMARGIN 1
#define TGMARGIN 3

void
draw_graph ()
{
  int minx, maxx, miny, maxy;

  minx = TMARGIN + font_height + TGMARGIN;
  maxx = width - GMARGIN - 1;
  miny = TMARGIN + font_height + TGMARGIN;
  maxy = height - GMARGIN - 1;

  draw_line (minx, miny, minx, maxy, -1);
  draw_line (minx, miny, maxx, miny, -1);
  draw_line (maxx, miny, maxx, maxy, -1);
  draw_line (minx, maxy, maxx, maxy, -1);

  fill (0, 0, width-1, miny-1, -1);
  fill (0, miny, minx-1, height-1, -1);
  fill (maxx+1, miny, width-1, height-1, -1);
  fill (minx, maxy+1, maxx, height-1, -1);

  minx ++;
  miny ++;
  maxx --;
  maxy --;
  fill (minx, miny, maxx, maxy, 0);

  time_t min_time, max_time;
  int max_itemp, min_itemp;
  int i, z;

  min_time = events[0].time;
  max_time = events[num_events - 1].time;
  max_itemp = min_itemp = events[0].temp[0];

  for (i=0; i<num_events; i++)
    for (z=0; z<4; z++)
      {
	if (min_itemp > events[i].temp[z])
	  min_itemp = events[i].temp[z];
	if (max_itemp < events[i].temp[z])
	  max_itemp = events[i].temp[z];
      }

  // Pad out to whole hours
  min_time -= min_time % HOUR;
  max_time += 60*60-1;
  max_time -= max_time % HOUR;

  // Pad out to ten degree marks
  min_itemp += 20000;
  min_itemp -= min_itemp % 1000;
  min_itemp -= 20000;
  max_itemp += 20999;
  max_itemp -= max_itemp % 1000;
  max_itemp -= 20000;

  double temp_scale = (double)(maxy - miny + 1) / (max_itemp - min_itemp + 1);
  double time_scale = (double)(maxx - minx + 1) / (max_time - min_time + 1);

  struct tm *tm = localtime(&min_time);
  tzone = (min_time % DAY) - (tm->tm_hour * HOUR + tm->tm_min * 60 + tm->tm_sec) + DAY;

  int last_time = minx;
  for (i=1; i<num_events; i++)
    {
      int this_time = (int)((events[i].time - min_time) * time_scale + minx);
      int base_time = last_time;
      int color;

      if (events[i].furnace.heat)
	color = 0x320000;
      else if (events[i].furnace.humidify)
	color = 0x323200;
      else if (events[i].furnace.cool)
	color = 0x000032;
      else if (events[i].furnace.fan)
	color = 0x003200;
      else
	{
	  int h = ((events[i].time-tzone)%DAY) / HOUR;
	  if (h < 4 || h >= 20)
	    color = 0;
	  else if (h >= 8 && h < 16)
	    color = 0x001020;
	  else
	    {
	      double daylight = sin(((events[i].time-tzone-6*HOUR)%DAY)*M_PI*6/DAY);
	      daylight = (daylight + 1)/2;
	      color = (int)(daylight * 16) * 256 + (int)(daylight * 32);
	    }
	  base_time++;
	}
      if (base_time <= this_time)
	fill (base_time, miny, this_time, maxy, color);

      last_time = this_time;
    }

  char buf[10];
  time_t t;

  int dt = HOUR;
  if (ndays > 1)
    dt = HOUR * 3;
  if (ndays > 4)
    dt = HOUR * 6;
  if (ndays > 8)
    dt = DAY;
  for (t=daystart(min_time); t<max_time; t += dt)
    {
      int this_time = (int)((t - min_time) * time_scale + minx);
      if (t <= min_time)
	continue;
      tm = localtime(&t);
      int color;
      if (tm->tm_hour < 6 || tm->tm_hour > 18)
	color = 0x1e1e1e;
      else if (tm->tm_hour == 6 || tm->tm_hour == 18)
	color = 0x2d2d2d;
      else
	color = 0x3c3c3c;
      draw_line(this_time, miny, this_time, maxy, color);
    }

  tm = localtime(&min_time);

  if (ndays > 2)
    {
      for (t=daystart(min_time); t<max_time; t += DAY)
	{
	  int this_time = (int)((t - min_time) * time_scale + minx);
	  if (t < min_time)
	    continue;
	  tm = localtime(&t);
	  sprintf(buf, "%s%d", month_names[tm->tm_mon], tm->tm_mday);
	  if (this_time > width - strlen(buf)*6-1)
	    continue;
	  draw_string (this_time, TMARGIN, 0, buf, 0);
	}
    }
  else
    {
      int dt = (ndays > 1) ? 6 : 3;
      while (dt < 12 && dt*HOUR*time_scale < 6*4)
	dt *= 2;
      sprintf(buf, "%s%d", month_names[tm->tm_mon], tm->tm_mday);
      draw_string (minx, TMARGIN, 0, buf, 0);
      for (t=daystart(min_time); t<max_time; t += dt*HOUR)
	{
	  int this_time = (int)((t - min_time) * time_scale + minx);
	  if (this_time < minx + 36)
	    continue;
	  tm = localtime(&t);
	  sprintf(buf, "%d%s", (tm->tm_hour+11)%12+1, tm->tm_hour<12 ? "am" : "pm");
	  if (this_time > width - strlen(buf)*3-1)
	    continue;
	  draw_string (this_time - 3*strlen(buf), TMARGIN, 0, buf, 0);
	}
    }

  for (int temp = min_itemp; temp <= max_itemp; temp += 1000)
    {
      int this_temp = (int)((temp - min_itemp) * temp_scale + miny);
      sprintf(buf, "%d%c", temp/100, 128);
      int ty = this_temp-3*(strlen(buf)-1);
      if (ty > height - 6*strlen(buf) - 1)
	ty = height - 6*strlen(buf) - 1;
      draw_string (minx - TGMARGIN, ty, 3, buf, 0);
      if (temp < max_itemp && temp > min_itemp)
	draw_line (minx+1, this_temp, maxx-1, this_temp, 0x1e1e1e);
    }

  for (z=3; z>=0; z--)
    {
      int last_temp = (int)((events[0].temp[z] - min_itemp) * temp_scale + miny);
      static int colors[] = { 0xf00000, 0x00f000, 0x0000f0, 0x808000 };
      int color = colors[z];
      int last_x = (int)((events[0].time - min_time) * time_scale + minx);
      
      for (i=1; i<num_events; i++)
      {
	int this_temp = (int)((events[i].temp[z] - min_itemp) * temp_scale + miny);
	int this_time = (int)((events[i].time - min_time) * time_scale + minx);

	if (events[i].time > events[i-1].time+10*60)
	  set_op (this_time, this_temp, color);
	else
	  draw_oline (last_x, last_temp, this_time, this_temp, color);
	last_x = this_time;
	last_temp = this_temp;
      }
    }
}

/* ====================================================================== */

int
main(int argc, char **argv)
{
  FILE *png = fopen("graphit.png", "wb");
  if (!png)
    {
      perror("graphit.png");
      exit(1);
    }

  while (argc > 1)
    {
      if (strcmp (argv[1], "-g") == 0 && argc > 2)
	{
	  sscanf(argv[2], "%d%*c%d", &width, &height);
	  argc -= 2;
	  argv += 2;
	}
      else if (strcmp (argv[1], "-d") == 0 && argc > 2)
	{
	  sscanf(argv[2], "%d%*c%d", &ndays, &bdays);
	  argc -= 2;
	  argv += 2;
	}
      else
	{
	  argc --;
	  argv ++;
	}
    }

  int x, y;
  rgb = (png_byte **) malloc (height * sizeof (png_byte *));
  for (y=0; y<height; y++)
    rgb[y] = (png_byte *) malloc (width * 3);

  chdir("/envy/dj/house/furnace/sw/logs");
  chdir("/root/src/sw/logs");

  time_t now;
  time(&now);

  max_events = 300;
  events = (LogEvent *) malloc (max_events * sizeof(LogEvent));

  for (int d=0; d<ndays; d++)
    {
      FILE *today = openlog(now - DAY * d - DAY * bdays);
      if (today)
	{
	  snarf_log_events(today);
	  fclose(today);
	}
    }

  qsort (events, num_events, sizeof(LogEvent), log_sort);

#if 0
  for (int i=0; i<num_events; i++)
    {
      LogEvent le = events[i];
      struct tm *tm = localtime((time_t *)&le.time);
      printf("%02d-%02d %02d:%02d:%02d ", tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
      zone(0, le.temp[0], le.humidity[0], le.furnace.zone0);
      zone(1, le.temp[1], le.humidity[1], le.furnace.zone1);
      zone(2, le.temp[2], le.humidity[2], le.furnace.zone2);
      zone(3, le.temp[3], le.humidity[3], le.furnace.zone3);
      printf("%c %c %c %c\n",
	     le.furnace.heat ? 'H' : '-',
	     le.furnace.cool ? 'C' : '-',
	     le.furnace.fan ? 'F' : '-',
	     le.furnace.humidify ? 'W' : '-');
    }
#endif

  for (x=0; x<width; x++)
    for (y=0; y<height; y++)
      set_p(x, y, (x*y)*(x^y));

  /* Draw graph here */
  draw_graph();

  png_byte **rgb2 = (png_byte **) malloc (height * sizeof (png_byte *));
  for (y=0; y<height; y++)
    rgb2[y] = rgb[height-y-1];

  png_structp png_ptr = png_create_write_struct
    (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
    exit(1);
  
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
    {
      png_destroy_write_struct(&png_ptr,
			       (png_infopp)NULL);
      exit(1);
    }

  if (setjmp(png_jmpbuf(png_ptr)))
    exit(1);
  png_init_io(png_ptr, png);

  png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB,
	       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_set_rows(png_ptr, info_ptr, rgb2);
  png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, 0);

  exit(0);
}
