// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "msgtypes.h"
#include "msghdr.h"

#define DEBUG 0

// This is the socket we listen on for new connections
int mux_fd = -1;

//-----------------------------------------------------------------------------

struct Message {
  int locks;
  int length;
  char *data;
};

struct Queue {
  int max, count;
  int put, get;
  Message **msg;
  int bufp;
  Queue();
  ~Queue();
  void add(Message *);
  void del();
  void write(int fd);
};

struct Connection {
  char len_buf[sizeof(MH_LengthType)];
  char *buf;
  int snooping;
  int bufp;
  int bufsize;
  int fd;
  Queue write_queue;

  Connection(int fd);
  ~Connection();

  int read_some_once(int first_time);
  void read_some();
  void write_some();
};

//-----------------------------------------------------------------------------

Queue::Queue()
{
  max = 10;
  count = 0;
  put = get = 0;
  msg = new Message *[10];
  bufp = 0;
}

Queue::~Queue()
{
  while (count)
    del();
  delete msg;
}

void
Queue::add(Message *m)
{
  m->locks ++;
  if (count == max)
    {
      int newmax = max + 10;
      int i;
      Message **msg2 = new Message *[newmax];
      for (i=0; i<count; i++)
	{
	  msg2[i] = msg[get++];
	  get %= max;
	}
      delete msg;
      msg = msg2;
      get = 0;
      put = count;
      max = newmax;
    }
  msg[put++] = m;
  put %= max;
  count ++;
}

void
Queue::del()
{
  if (count == 0)
    return;
  Message *m = msg[get++];
  get %= max;
  if (--m->locks == 0)
    {
      delete m->data;
      delete m;
    }
  count --;
}

void
Queue::write(int fd)
{
  while (count)
    {
      Message *m = msg[get];
      int nbytes = m->length - bufp;
      int written = ::write(fd, m->data + bufp, nbytes);
      if (written <= 0)
	return;

      if (written < nbytes)
	{
	  bufp += written;
	  continue;
	}

      bufp = 0;
      del();
    }
}

//-----------------------------------------------------------------------------

// Index is fd
Connection **connections = 0;
int max_connections = 0;

Connection::Connection(int new_fd)
  :write_queue()
{
  fd = new_fd;
  if (connections == 0)
    {
      max_connections = fd + 5;
      connections = (Connection **) malloc (max_connections * sizeof (Connection *));
    }
  else if (max_connections <= fd)
    {
      int new_max = fd + 5;
      connections = (Connection **) realloc (connections, new_max * sizeof (Connection *));
      memset (connections + max_connections,
	      0,
	      (new_max - max_connections) * sizeof (Connection *));
      max_connections = new_max;
    }
  connections[fd] = this;

  bufsize = sizeof(len_buf);
  buf = len_buf;
  bufp = 0;

  Message *intro_msg = new Message;
  intro_msg->locks = 0;
  intro_msg->length = sizeof(MessageHeader) + 4;
  intro_msg->data = new char[sizeof(MessageHeader)];
  strcpy (intro_msg->data+sizeof(MessageHeader), "mux");
  MessageHeader *mh = (MessageHeader *)(intro_msg->data);
  mh->length = 4;
  mh->dest_id = fd;
  mh->source_id = 1;
  mh->type = MT_my_name;
  write_queue.add(intro_msg);

  snooping = 0;
}

Connection::~Connection()
{
#if DEBUG
  printf("deleting connection %d\n", fd);
#endif
  connections[fd] = 0;
  close (fd);
}

void
Connection::write_some()
{
  write_queue.write(fd);
}

static void
setcolor (int nc)
{
  static int oldcolor = 0;
  if (nc != oldcolor)
    {
      oldcolor = nc;
      if (nc)
	printf("\033[%dm", 30+nc);
      else
	printf("\033[0m");
    }
}

static void
dump(const void *vbuf, int len)
{
  const unsigned char *buf = (const unsigned char *)vbuf;
  int col = 0;
  const MessageHeader *mh = (const MessageHeader *)buf;
  printf("%4d %02x->%02x %04x",
	 mh->length, mh->source_id, mh->dest_id, mh->type);
  buf += sizeof(MessageHeader);
  len -= sizeof(MessageHeader);
  col = 0;
  while (len--)
    {
      if (col >= 75)
	{
	  printf("\n\t\t");
	  col = 0;
	}
      unsigned char c = (*buf++) & 0xff;

      if (isgraph(c))
	{
	  setcolor(0);
	  printf(" %c", c);
	  col += 2;
	}
      else
	{
	  static int toggle = 0;
	  toggle = 1-toggle;
	  setcolor(toggle ? 3 : 6);
	  printf(" %02x", c);
	  col += 3;
	}
    }
  setcolor(0);
  printf("\n");
}

static void
send_snoop(Message *m, int except_id)
{
  for (int id=0; id<max_connections; id++)
    if (id != except_id
	&& connections[id]
	&& connections[id]->snooping)
      {
#if DEBUG
	printf("SEND: snoop to %d\n", id);
#endif
	connections[id]->write_queue.add(m);
      }
}

int
Connection::read_some_once(int first_time)
{
  int nread;
  int msg_length;

  if (bufp >= bufsize || bufp < 0)
    abort();

  errno = 0;
  nread = ::read(fd, buf + bufp, bufsize - bufp);
  if (nread <= 0)
    {
      if (first_time)
	{
	  // connection is closed.
	  delete this;
	}
      return 0;
    }
  bufp += nread;

  if (bufp < sizeof(MH_LengthType))
    return 1;

  msg_length = *(MH_LengthType *)buf + sizeof(MessageHeader);

  if (buf == len_buf)
    {
      char *newbuf = new char[msg_length];
      memcpy (newbuf, buf, bufp);
      buf = newbuf;
      bufsize = msg_length;
#if DEBUG
      printf("\nincoming message: %d bytes\n", msg_length);
#endif
      // Maybe we can read the rest of the message now.
      nread = ::read(fd, buf + bufp, msg_length - bufp);
      bufp += nread;
    }

  if (bufp < msg_length)
    return 1;

#if DEBUG
  printf("\033[32mbuilding message\033[0m\n");
  dump(buf, bufp);
#endif
  Message *m = new Message;
  m->locks = 0;
  m->length = msg_length;
  m->data = buf;

  buf = len_buf;
  bufsize = sizeof(len_buf);
  bufp = 0;

  MessageHeader *mh = (MessageHeader *)(m->data);
  mh->source_id = fd;

  if (mh->dest_id == 1)
    {
      switch (mh->type)
	{
	case MT_snoop:
	  connections[fd]->snooping = m->data[sizeof(MessageHeader)];
#if DEBUG
	  printf("SNOOP: fd %d snoop %d\n", fd, connections[fd]->snooping);
#endif
	  break;
	}
      send_snoop(m, mh->dest_id);
      if (m->locks == 0)
	{
	  delete m->data;
	  delete m;
	}
      return 1;
    }

  if (mh->dest_id)
    {
      if (mh->dest_id < max_connections
	  && connections[mh->dest_id])
	{
#if DEBUG
	  printf("SEND: direct to %d\n", mh->dest_id);
#endif
	  connections[mh->dest_id]->write_queue.add(m);
	}
      send_snoop(m, mh->dest_id);
      return 1;
    }

  for (int id=0; id<max_connections; id++)
    if (id != fd && connections[id])
      {
#if DEBUG
	printf("SEND: multi to %d\n", id);
#endif
	connections[id]->write_queue.add(m);
      }

  if (m->locks == 0)
    {
      delete m->data;
      delete m;
    }
  return 1;
}

void
Connection::read_some()
{
  int first_time = 1;
  while (read_some_once(first_time))
    first_time = 0;
}

//-----------------------------------------------------------------------------

int
set_up_socket_once (void)
{
  int one = 1;
  struct sockaddr_in addr;

  mux_fd = socket (PF_INET, SOCK_STREAM, 0);
  setsockopt (mux_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

  addr.sin_family = AF_INET;
  addr.sin_port = htons(MUX_PORT);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind (mux_fd, (struct sockaddr *)&addr, sizeof(addr)))
    {
      close (mux_fd);
      return 0;
    }

  if (listen (mux_fd, 20))
    {
      close (mux_fd);
      return 0;
    }

  setsockopt (mux_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

  return 1;
}

void
set_up_socket (void)
{
  int i;
  for (i=0; i<6; i++)
    if (set_up_socket_once ())
      return;
  printf("unable to set up mux listen socket.\n");
  exit(1);
}

void
handle_event (void)
{
  fd_set read_fds;
  fd_set write_fds;
  fd_set except_fds;
  int max_fd;
  int i;

  FD_ZERO (&read_fds);
  FD_ZERO (&write_fds);
  FD_ZERO (&except_fds);

  FD_SET (mux_fd, &read_fds);
  max_fd = mux_fd + 1;

  for (i=0; i<max_connections; i++)
    if (connections[i])
      {
	FD_SET(i, &read_fds);
	if (max_fd <= i)
	  max_fd = i+1;
	if (connections[i]->write_queue.count)
	  FD_SET(i, &write_fds);
      }

  i = select (max_fd, &read_fds, &write_fds, &except_fds, 0);

  if (FD_ISSET (mux_fd, &read_fds))
    {
      int one = 1;
      int new_fd = accept (mux_fd, 0, 0);
      setsockopt (mux_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
      setsockopt (mux_fd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
      fcntl(new_fd, F_SETFL, O_NONBLOCK);
#if DEBUG
      printf("adding connection %d\n", new_fd);
#endif
      new Connection(new_fd);
    }

  for (i=0; i<max_fd; i++)
    {
      if (!connections[i])
	continue;
      if (FD_ISSET(i, &read_fds))
	connections[i]->read_some();
      if (!connections[i])
	continue;
      if (FD_ISSET(i, &write_fds))
	connections[i]->write_some();
    }
}

int
main(int argc, char **argv)
{
  // Don't close stdout, because we use fd 1 to mean "us"
  setbuf(stdout, 0);

  while (1)
    {
      set_up_socket ();
      printf("mux ready.\n");
      while (1)
	handle_event ();
    }
}
