// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>

#include "msg.h"
#include "vars.h"
#include "misc.h"
#include "furnace.h"

#include "eedigits.h"

#define TIMEOUT_MS (BUTTON_TIMEOUT * 1000)
#define REPEAT_MS 350

#define MODE_HEAT     'N'
#define MODE_FAN      'H'
#define MODE_COOL     'L'
#define MODE_HUMIDIFY 'J'
#define MODE_OFF      'R'

#if 0
#define DMODE(x) printf("Mode: %s\n", #x);
#else
#define DMODE(x)
#endif

int myzone;

int pending_alarm = 0;

Var heating("heating-mode");
Var cooling("cooling-mode");
Var autocool("auto-cool-temp");
Var autoheat("auto-heat-temp");

Var max_heat, min_heat;
Var max_cool, min_cool;
Var max_humidity, min_humidity;

FurnaceData furnace, old_furnace;
double outside_temp, outside_humidity;
double my_temp, my_humidity;

//-----------------------------------------------------------------------------

static char lbuf[1024];
static char *lb = lbuf+1;

static void
lput(char c, int rep=1)
{
  while (rep--)
    *lb++ = c;
}
static void
lput(const unsigned char *c, int l)
{
  memcpy (lb, (void *)c, l);
  lb += l;
}
static void
lput(char *c)
{
  int l = strlen (c);
  memcpy (lb, c, l);
  lb += l;
}
static void
lprf(const char *fmt, ...)
{
  va_list a;
  va_start (a, fmt);
  vsprintf(lb, fmt, a);
  lb += strlen(lb);
  va_end(a);
}
static void
lput(void)
{
  lbuf[0] = myzone;
  Message l(lbuf, lb-lbuf, MT_lcd);
  l.send();
  usleep(l.length*1041);
  lb = lbuf + 1;
}

static void
lgoto(int row, int col)
{
  lput(16);
  lput(64+row*20+col);
}

static void
lpause(int msec)
{
  short x = msec;
  lput();
  Message p(&x, sizeof(x), MT_lcd_pause);
  p.send();
  usleep(msec*1000);
}

static void
lclear (void)
{
  lput("\033R@\014");
  lpause(20);
}

//-----------------------------------------------------------------------------

void
show_icon (int row, int col, int mode)
{
  lput("\033FH");
  lgoto(row,col);
  lput(mode);
  lput(mode+1);
  lgoto(row+1,col);
  lput(mode+20);
  lput(mode+21);
  lput("\033F@");
  lput();
}

//-----------------------------------------------------------------------------

void
download_eeprom_page (const unsigned char *data, int pageno)
{
  char buf[109 + 120*4];
  char *bp = buf;

  lclear();
  lput("\033DG");
  lpause(200);
  lput(data, 120*4);
  lpause(200);
  lput("\033X");
  lput(64+pageno);
  lpause(100);
}

char big_print_cache[4];

void
big_print_restart()
{
  memset(big_print_cache, 0, 4);
}

void
big_print(double x, char F='F')
{
  char buf[20];
  int r, d, c, d1, d2;
  int need_fontset = 0;

  if (x >= 0 && x < 100)
    sprintf(buf, "%2d", (int)x);
  else
    strcpy (buf, "  ");

  d1 = (buf[0] == big_print_cache[0]) ? 1 : 0;
  d2 = (buf[1] == big_print_cache[1]) ? 0 : 1;

  if (d1 <= d2)
    {
      need_fontset = 1;
      lput("\033FD");
      for (r=0; r<4; r++)
	{
	  lgoto(r, 6 + d1*4);
	  for (d=d1; d<=d2; d++)
	    for (c=0; c<4; c++)
	      {
		int dig = buf[d] - '0';
		if (buf[d] == ' ')
		  lput('\'');
		else
		  lput(32 + (dig%5)*4 + (dig/5)*80 + r*20 + c);
	      }
	}
    }
  if (d2 == 0)
    lgoto (3, 14);
  d = (int)(x*10)%10 + '0';
  if (need_fontset)
    lput("\033F@");
  if (d != big_print_cache[2])
    {
      lprf(".%c", d);
    }
  if (F != big_print_cache[3])
    {
      lgoto(0, 14);
      if (F != '%')
	lput('\200');
      lput(F);
    }
  lput();
  big_print_cache[0] = buf[0];
  big_print_cache[1] = buf[1];
  big_print_cache[2] = d;
  big_print_cache[3] = F;
}

void
small_print (int row, int col, double x, int *cache=0, char suff=0)
{
  char buf[20], *c;
  if (suff)
    {
      int xd = (int)(x+0.5);
      if (cache)
	{
	  if (*cache == xd)
	    return;
	  *cache = xd;
	}
      lgoto(row, col);
      lprf("%3d%c", xd, suff);
    }
  else
    {
      int xd = (int)(x*10+0.5);
      if (cache)
	{
	  if (*cache == xd)
	    return;
	  *cache = xd;
	}
      sprintf(buf, "%5.1f", xd/10.0);
      c = strchr(buf, '.');
      *c++ = 0;
      lgoto(row, col);
      lput(buf);
      lput(*c + 132 - '0');
    }
  lput();
}

//-----------------------------------------------------------------------------

static jmp_buf back_to_idle_mode;
int repeat_ok = 0;

static void
idle()
{
  Message::set_alarm();
  longjmp(back_to_idle_mode,1);
}

/* Note that up..right must correspond to the button numbers returned
   by Msg.  */
typedef enum {
  M_Up,
  M_Down,
  M_Left,
  M_Right,
  M_Climate,
  M_Furnace,
  M_Variable,
  M_NoEvent
} MEvent;

static const char *MNames[] = {
  "up", "down", "left", "right", "climate", "furnace", "variable", "no-event" };

static MEvent pending_event = M_NoEvent;

void redo_event (MEvent e)
{
  pending_event = e;
}

static int
same(double a, double b)
{
  if ((int)(a*10+0.5) == (int)(b*10+0.5))
    return 1;
  return 0;
}

static MEvent
next_event ()
{
  static time_t last_button_time = 0;
  static int repeat = 0;
  static MEvent last_button;

  if (pending_event != M_NoEvent)
    {
      MEvent rv = pending_event;
      pending_event = M_NoEvent;
      return rv;
    }

  while (1)
    {
      if (pending_alarm)
	{
	  Message::set_alarm(pending_alarm);
	  pending_alarm = 0;
	}
      Message *m = Message::next();

      if (!m)
	{
	  if (repeat)
	    {
	      pending_alarm = REPEAT_MS;
	      return last_button;
	    }
	  else
	    idle();
	}

      switch (m->type)
	{
	case MT_set_furnace:
	  {
	    if (memcmp (&furnace, m->data, sizeof(FurnaceData)))
	      {
		old_furnace = furnace;
		furnace = *(FurnaceData *)m->data;
		return M_Furnace;
	      }
	    break;
	  }
	case MT_climate:
	  {
	    ClimateData *cd = (ClimateData *)m->data;
	    if (cd->zone == 3)
	      {
		if (same(outside_temp, cd->temperature)
		    && same(outside_humidity, cd->humidity))
		  break;
		outside_temp = cd->temperature;
		outside_humidity = cd->humidity;
		return M_Climate;
	      }
	    else if (cd->zone == myzone)
	      {
		if (same(my_temp, cd->temperature)
		    && same(my_humidity, cd->humidity))
		  break;
		my_temp = cd->temperature;
		my_humidity = cd->humidity;
		return M_Climate;
	      }
	    break;
	  }
	case MT_button:
	  {
	    ButtonData *b = (ButtonData *)m->data;
	    if (b->zone == myzone)
	      {
		if (b->pressed)
		  {
		    last_button = (MEvent)b->button;
		    if (repeat_ok
			&& (last_button == M_Up || last_button == M_Down))
		      {
			pending_alarm = REPEAT_MS * 2;
			repeat = 1;
		      }
		    else
		      {
			Message::set_alarm();
		      }
		    return (MEvent)b->button;
		  }
		else
		  {
		    pending_alarm = TIMEOUT_MS;
		    repeat = 0;
		  }
	      }
	    break;
	  }
	}
    }
}

//-----------------------------------------------------------------------------

typedef struct {
  char *text;
  void (*function)(void);
} MenuEntry;

struct Menu {
  MenuEntry *ents;
  int n_ents;
  int top_ent;
  int selected;
  Menu (MenuEntry *);
};

Menu::Menu(MenuEntry *e)
{
  ents = e;
  for (n_ents = 0; e[n_ents].text; n_ents++)
    ;
  top_ent = 0;
  selected = 0;
}

static void
do_menu (Menu &m, int return_button)
{
  int need_refresh = 1;
  int old_top = -1, old_sel = -1;

  DMODE(menu);

  while (1)
    {
      if (m.selected < 0)
	m.selected = 0;
      if (m.selected >= m.n_ents)
	m.selected = m.n_ents - 1;
      if (m.n_ents > 4)
	{
	  if (m.selected < m.top_ent)
	    m.top_ent = m.selected;
	  if (m.selected > m.top_ent+3)
	    m.top_ent = m.selected - 3;
	}
      if (old_top != m.top_ent)
	{
	  int row;
	  //printf("menu: referesh top %d sel %d\n", m.top_ent, m.selected);
	  lprf("\033R@");
	  lput(12);
	  lput(1);
	  for (row=m.top_ent; row<m.top_ent+4 && row < m.n_ents; row++)
	    lprf("%-20.20s", m.ents[row].text);
	  for (; row<m.top_ent+4; row++)
	    lprf("%-20.20s", "");
	  old_top = m.top_ent;
	  old_sel = -1;
	}
      if (old_sel != m.selected)
	{
	  int i = 0;
	  i ^= 1 << (m.selected - m.top_ent);
	  //printf("menu: toggle rows %d\n", i);
	  lprf("\033R%c", 64+i);
	  lput();
	  old_sel = m.selected;
	}

      int e = next_event();
      // if (e != M_Furnace && e != M_Climate)
      // printf("menu: event %s\n", MNames[e]);
      switch (e)
	{
	case M_Up:
	  m.selected --;
	  break;
	case M_Down:
	  m.selected ++;
	  break;
	case M_Left:
	case M_Right:
	  if (e == return_button)
	    {
	      lput("\033R@");
	      return;
	    }
	  if (m.ents[m.selected].function)
	    {
	      m.ents[m.selected].function();
	      old_top = -1;
	    }
	  break;
	}
    }
}

//-----------------------------------------------------------------------------

int ssc_mytemp=-1, ssc_otemp=-1;

void
set_show_climate(int refresh=0)
{
  if (refresh)
    ssc_mytemp = ssc_otemp = -1;
  small_print(2, 0, my_temp, &ssc_mytemp);
  small_print(3, 0, outside_temp, &ssc_otemp);
}

//-----------------------------------------------------------------------------

static void
set_humidity_refresh()
{
  lclear();
  lput("\016Set\rHumidity");
  big_print_restart();
  big_print((max_humidity + min_humidity)/2, '%');
  set_show_climate(1);
}

static void
set_humidity(void)
{
  double humidity = 0, diff = 0;
  DMODE(set_humidity);

  humidity = (min_humidity + max_humidity)/2;
  diff = max_humidity - min_humidity;
  set_humidity_refresh();

  while (1)
    {
      int e = next_event();
      switch (e)
	{
	case M_Climate:
	  set_show_climate();
	  break;
	case M_Up:
	case M_Down:
	  if (e == M_Up)
	    humidity += 1;
	  else
	    humidity -= 1;
	  if (humidity <= 1)
	    humidity = 1;
	  if (humidity >= 99)
	    humidity = 99;
	  min_humidity.set(humidity-diff/2);
	  max_humidity.set(humidity+diff/2);
	  big_print((max_humidity + min_humidity)/2, '%');
	  break;
	case M_Left:
	  return;
	}
    }
}

//-----------------------------------------------------------------------------

static void
set_temp_range_refresh(Var *min, Var *max)
{
  lclear();
  lput("\016Set\rRange");
  if (!min)
    {
      lgoto(1,7);
      lput("Furnace");
      lgoto(2,7);
      lput("is off");
      lput();
    }
  else
    {
      big_print_restart();
      big_print((double)*max - (double)*min, ' ');
      small_print(0, 16, (double)*max + 0.01);
      small_print(3, 16, (double)*min + 0.01);
    }      
  
  set_show_climate(1);
}

static void
set_temp_range(void)
{
  Var *min=0, *max=0;
  double temp = 0, diff = 0;
  DMODE(set_range);

  if (heating)
    min = &min_heat, max = &max_heat;
  else if (cooling)
    min = &min_cool, max = &max_cool;

  if (min)
    {
      temp = (*min + *max)/2;
      diff = *max - *min;
    }
  set_temp_range_refresh(min, max);

  while (1)
    {
      int e = next_event();
      switch (e)
	{
	case M_Climate:
	  set_show_climate();
	  break;
	case M_Up:
	case M_Down:
	  if (!min)
	    break;
	  if (e == M_Up)
	    diff += 1;
	  else
	    diff -= 1;
	  if (diff <= 0)
	    diff = 1;
	  min->set(temp-diff/2);
	  max->set(temp+diff/2);
	  set_temp_range_refresh(min, max);
	  break;
	case M_Left:
	  return;
	}
    }
}

//-----------------------------------------------------------------------------

static void
set_autotemp_refresh(Var *var, char *typestring, int mode)
{
  lclear();
  lput("\016Auto\r");
  lput(typestring);
  big_print_restart();
  big_print(var->value);
  show_icon (0, 17, mode);
  set_show_climate(1);
}

static void
set_autotemp(Var *var, char *typestring, int mode)
{
  double temp = var->value;
  DMODE(set_autotemp);

  set_autotemp_refresh(var, typestring, mode);

  while (1)
    {
      int e = next_event();
      switch (e)
	{
	case M_Climate:
	  set_show_climate();
	  break;
	case M_Up:
	case M_Down:
	  if (e == M_Up)
	    temp += 1;
	  else
	    temp -= 1;
	  if (temp <= 50)
	    temp = 50;
	  if (temp >= 90)
	    temp = 90;
	  var->set(temp);
	  big_print(var->value);
	  break;
	case M_Left:
	  return;
	}
    }
}

static void set_autocool(void) { set_autotemp(&autocool, "Cool", MODE_COOL); }
static void set_autoheat(void) { set_autotemp(&autoheat, "Heat", MODE_HEAT); }

//-----------------------------------------------------------------------------

static MenuEntry set_temp_left_menu_ents[] = {
  // ----+----1----+----2
  { "Show Zones", 0 },
  { "Zone History", 0 },
  { "Outside Weather", 0 },
  { "Outside History", 0 },
  { "Furnace Activity", 0 },
  { "Weather Report", 0 },
  { 0, 0 }
};
static Menu set_temp_left_menu(set_temp_left_menu_ents);

static MenuEntry set_temp_right_menu_ents[] = {
  // ----+----1----+----2
  { "Set Humidity", set_humidity },
  { "Set Temp Range", set_temp_range },
  { "Set AutoCool Temp", set_autocool },
  { "Set AutoHeat Temp", set_autoheat },
  { 0, 0 }
};
static Menu set_temp_right_menu (set_temp_right_menu_ents);

void
set_temp_show(Var *min, Var *max)
{
  if (min)
    {
      big_print((*min + *max + 0.01) / 2);
      small_print(0, 16, (double)*max + 0.01);
      small_print(3, 16, (double)*min + 0.01);
    }
}

static void
set_temp_refresh(Var *min, Var *max, char *mode_string)
{
  lclear();
  lput("\016");
  lput(mode_string);
  lput("\rTemp");
  if (!min)
    {
      lgoto(1,7);
      lput("Furnace");
      lgoto(2,7);
      lput("is off");
      lput();
    }
  big_print_restart();
  set_temp_show(min, max);
  set_show_climate(1);
}

void
set_temp_mode()
{
  Var *min=0, *max=0;
  double temp = 0, diff = 0;
  DMODE(set_temp);
  char *mode_string = "off";

  repeat_ok = 1;
  big_print_restart();

  if (heating)
    {
      min = &min_heat;
      max = &max_heat;
      mode_string = "Heat";
    }
  else if (cooling)
    {
      min = &min_cool;
      max = &max_cool;
      mode_string = "Cool";
    }

  if (min)
    {
      temp = (*min + *max)/2;
      diff = *max - *min;
    }
  if (pending_event != M_Left
      && pending_event != M_Right)
    set_temp_refresh(min, max, mode_string);

  while (1)
    {
      int e = next_event();
      // printf("set temp: e = %d %s\n", e, MNames[e]);
      switch (e)
	{
	case M_Climate:
	  set_show_climate();
	  break;
	case M_Up:
	case M_Down:
	  if (!min)
	    break;
	  if (e == M_Up)
	    temp += 0.5;
	  else
	    temp -= 0.5;
	  min->set(temp-diff/2);
	  max->set(temp+diff/2);
	  set_temp_show(min, max);
	  break;
	case M_Left:
	  do_menu (set_temp_left_menu, M_Right);
	  DMODE(set_temp_lm);
	  set_temp_refresh(min, max, mode_string);
	  break;
	case M_Right:
	  do_menu (set_temp_right_menu, M_Left);
	  DMODE(set_temp_rm);
	  set_temp_refresh(min, max, mode_string);
	  break;
	}
    }
}

//-----------------------------------------------------------------------------

static void
idle_show_furnace ()
{

  int mode1, mode2;
  if (furnace.zone[myzone] == 0)
    mode1 = mode2 = MODE_OFF;
  else if (furnace.humidify)
    mode1 = MODE_HEAT, mode2 = MODE_HUMIDIFY;
  else if (furnace.heat)
    mode1 = MODE_HEAT, mode2 = furnace.fan ? MODE_FAN : MODE_OFF;
  else if (furnace.cool)
    mode1 = MODE_COOL, mode2 = furnace.fan ? MODE_FAN : MODE_OFF;
  else if (furnace.fan)
    mode1 = MODE_FAN, mode2 = MODE_OFF;
  else
    mode1 = mode2 = MODE_OFF;

  show_icon (0, 17, mode2);
  show_icon (2, 17, mode1);
}

int small_humid=-1, small_otemp=-1, small_ohumid=-1;

static void
idle_climate(int refresh=0)
{
  if (refresh)
    {
      small_humid = small_otemp = small_ohumid = -1;
      big_print_restart();
    }
  big_print(my_temp);
  small_print(0, 0, my_humidity, &small_humid, '%');
  small_print(2, 0, outside_temp, &small_otemp, 'F');
  small_print(3, 0, outside_humidity, &small_ohumid, '%');
}

static void
exit_proc(int)
{
  lput("\017\014");
  lput();
  exit(0);
}

int
main(int argc, char **argv)
{
  char namebuf[10];
  Message *m;

  if (argc < 2)
    {
      fprintf(stderr, "Usage: zoneui <zone0..2>\n");
      exit(1);
    }

  signal(SIGTERM, exit_proc);
  signal(SIGINT, exit_proc);

  myzone = atoi(argv[1]);

  if (argc > 2)
    Message::set_host(argv[2]);

  sprintf(namebuf, "zone%d", myzone);
  max_heat.set_name(myzone, "max-heat");
  min_heat.set_name(myzone, "min-heat");
  max_cool.set_name(myzone, "max-cool");
  min_cool.set_name(myzone, "min-cool");
  max_humidity.set_name(myzone, "max-humidity");
  min_humidity.set_name(myzone, "min-humidity");

  Message::set_my_name(namebuf);
  Message::init();

  Var::wait_for_all_ids();

  double last_temp = -999;
  double last_humidity = -999;

  download_eeprom_page (eedigits[0], 0);
  download_eeprom_page (eedigits[1], 1);
  download_eeprom_page (eedigits[2], 2);
  download_eeprom_page (eedigits[3], 3);
  download_eeprom_page (eedigits[4], 4);

  setjmp(back_to_idle_mode);
  DMODE(idle);

  lclear();
  lput("\033R@\001\003\017\033F@\033IA\033M@\033V@");
  idle_show_furnace();
  idle_climate(1);

  int light_on = 0;
  repeat_ok = 0;

  while (1)
    {
      MEvent e;
      switch (e = next_event())
	{
	case M_Furnace:
	case M_Climate:
	  idle_show_furnace ();
	  idle_climate();
	  break;

	case M_Up:
	case M_Down:
	case M_Left:
	case M_Right:
	  if (!light_on)
	    {
	      lput('\016');
	      lput();
	      light_on = 1;
	      repeat_ok = 1;
	    }
	  else
	    {
	      if (e == M_Left || e == M_Right)
		redo_event(e);
	      set_temp_mode();
	    }
	  break;
	}
    }
}
